from enum import Enum
from importlib.metadata import version

from pydantic import BaseModel

from dash_pydantic_form.fields import fields
from dash_pydantic_form.form_section import FormSection, Sections
from dash_pydantic_form.model_form import ModelForm
from dash_pydantic_form.utils import from_form_data, get_model_cls

BaseModel.__getitem__ = lambda self, key: self.__dict__.get(key)
BaseModel.to_plotly_json = lambda self: self.model_dump(mode="json")
Enum.to_plotly_json = lambda self: self.value

_css_dist = [
    {
        "relative_package_path": "pydf_styles.css",
        "namespace": "dash_pydantic_form",
    },
]
_js_dist = [
    {
        "relative_package_path": "pydf_clientside.js",
        "namespace": "dash_pydantic_form",
    },
    {
        "relative_package_path": "pydf_aggrid.js",
        "namespace": "dash_pydantic_form",
    },
]

__version__ = version(__package__)

__all__ = ["FormSection", "ModelForm", "Sections", "get_model_cls", "fields", "from_form_data", "__version__"]



from typing import get_args, get_origin

import pytest
from pydantic import BaseModel, Field

from dash_pydantic_form import utils


def test_ut0001_get_subitem():
    """Test get_subitem."""

    class Nested2(BaseModel):
        a: int

    class Nested(BaseModel):
        li: list[Nested2] = Field(default_factory=list)
        di: dict[str, Nested2] = Field(default_factory=dict)

    class Base(BaseModel):
        name: str
        x: Nested
        y: list[Nested] = Field(default_factory=list)
        z: dict[str, Nested] = Field(default_factory=dict)

    item = Base(
        name="test",
        x=Nested(li=[Nested2(a=1), Nested2(a=2)], di={"d0": Nested2(a=3)}),
        y=[
            Nested(li=[Nested2(a=4), Nested2(a=5)], di={"d1": Nested2(a=6)}),
            Nested(li=[Nested2(a=7), Nested2(a=8)], di={"d2": Nested2(a=9)}),
        ],
        z={
            "z0": Nested(li=[Nested2(a=10), Nested2(a=11)], di={"d3": Nested2(a=12)}),
            "z1": Nested(li=[Nested2(a=13), Nested2(a=14)], di={"d4": Nested2(a=15)}),
        },
    )

    assert utils.get_subitem(item, "name") == "test"
    assert utils.get_subitem(item, "x") == Nested(li=[Nested2(a=1), Nested2(a=2)], di={"d0": Nested2(a=3)})
    assert utils.get_subitem(item, "x:li") == [Nested2(a=1), Nested2(a=2)]
    assert utils.get_subitem(item, "x:li:1") == Nested2(a=2)
    assert utils.get_subitem(item, "x:di:d0:a") == 3  # noqa: PLR2004
    assert utils.get_subitem(item, "y:0:li:0:a") == 4  # noqa: PLR2004
    assert utils.get_subitem(item, "y:1:di:d2:a") == 9  # noqa: PLR2004
    assert utils.get_subitem(item, "z:z0:li:0:a") == 10  # noqa: PLR2004
    assert utils.get_subitem(item, "z:z1:di:d4:a") == 15  # noqa: PLR2004
    assert utils.get_subitem(item, "z:1:di:0:a") == 15  # noqa: PLR2004
    assert utils.get_subitem(item, "z:z2") is None

    with pytest.raises(AttributeError):
        utils.get_subitem(item, "u")
    with pytest.raises(AttributeError):
        utils.get_subitem(item, "z:z1:k")
    with pytest.raises(IndexError):
        utils.get_subitem(item, "z:3")
    with pytest.raises(IndexError):
        utils.get_subitem(item, "y:3")


def test_u0002_get_subitem_cls():
    """Test get_subitem_cls."""

    class Nested2(BaseModel):
        a: int

    class Nested(BaseModel):
        li: list[Nested2] = Field(default_factory=list)
        di: dict[str, Nested2] = Field(default_factory=dict)

    class Base(BaseModel):
        name: str
        x: Nested
        y: list[Nested] = Field(default_factory=list)
        z: dict[str, Nested] = Field(default_factory=dict)

    assert utils.get_subitem_cls(Base, "name") is str
    assert utils.get_subitem_cls(Base, "x") is Nested
    tmp = utils.get_subitem_cls(Base, "x:li")
    assert get_origin(tmp) is list
    assert get_args(tmp)[0] is Nested2
    assert utils.get_subitem_cls(Base, "x:li:0") is Nested2
    tmp = utils.get_subitem_cls(Base, "x:di")
    assert get_origin(tmp) is dict
    assert utils.get_subitem_cls(Base, "x:di:0") is Nested2
    assert utils.get_subitem_cls(Base, "x:di:d0") is Nested2
    assert utils.get_subitem_cls(Base, "z:somekey") is Nested
    assert utils.get_subitem_cls(Base, "z:somekey:di:otherkey") is Nested2
    assert utils.get_subitem_cls(Base, "z:0:di:1") is Nested2
    assert utils.get_subitem_cls(Base, "y:9") is Nested
    assert utils.get_subitem_cls(Base, "y:9:li:7") is Nested2

    with pytest.raises(AttributeError):
        utils.get_subitem(Base, "u")
    with pytest.raises(AttributeError):
        utils.get_subitem(Base, "x:u")
    with pytest.raises(AttributeError):
        utils.get_subitem(Base, "x:li:key")


def test_ut0003_get_model_value():
    """Test get_model_value."""

    class Nested2(BaseModel):
        a: int

    class Nested(BaseModel):
        li: list[Nested2] = Field(default_factory=list)
        di: dict[str, Nested2] = Field(default_factory=dict)

    class Base(BaseModel):
        name: str
        x: Nested
        y: list[Nested] = Field(default_factory=list)
        z: dict[str, Nested] = Field(default_factory=dict)

    item = Base(
        name="test",
        x=Nested(li=[Nested2(a=1), Nested2(a=2)], di={"d0": Nested2(a=3)}),
        y=[
            Nested(li=[Nested2(a=4), Nested2(a=5)], di={"d1": Nested2(a=6)}),
            Nested(li=[Nested2(a=7), Nested2(a=8)], di={"d2": Nested2(a=9)}),
        ],
        z={
            "z0": Nested(li=[Nested2(a=10), Nested2(a=11)], di={"d3": Nested2(a=12)}),
            "z1": Nested(li=[Nested2(a=13), Nested2(a=14)], di={"d4": Nested2(a=15)}),
        },
    )

    assert utils.get_model_value(item, "name", "") == "test"
    assert utils.get_model_value(item, "a", "x:li:0") == 1  # noqa: PLR2004
    assert utils.get_model_value(item, "a", "x:di:d0") == 3  # noqa: PLR2004
    assert utils.get_model_value(item, "a", "z:0:li:1") == 11  # noqa: PLR2004
    assert utils.get_model_value(item, "a", "x:li:3", True) is None
    assert utils.get_model_value(item, "a", "x:di:3", True) is None

    with pytest.raises(IndexError):
        assert utils.get_model_value(item, "a", "x:li:3", False)
    with pytest.raises(IndexError):
        assert utils.get_model_value(item, "a", "x:di:3", False)



import contextlib
import itertools
from copy import deepcopy
from functools import partial
from typing import Literal, Union

import dash_mantine_components as dmc
from dash import (
    ALL,
    MATCH,
    ClientsideFunction,
    Input,
    Output,
    State,
    callback,
    clientside_callback,
    ctx,
    dcc,
    html,
)
from dash.development.base_component import Component
from dash_iconify import DashIconify
from pydantic import BaseModel

from . import ids as common_ids
from .fields import BaseField, fields
from .form_section import Sections
from .utils import (
    SEP,
    Type,
    deep_merge,
    get_fullpath,
    get_model_cls,
    get_subitem,
    get_subitem_cls,
    handle_discriminated,
    model_construct_recursive,
)


def form_base_id(part: str, aio_id: str, form_id: str, parent: str = ""):
    """Form parts id."""
    return {"part": part, "aio_id": aio_id, "form_id": form_id, "parent": parent}


Children_ = Component | str | int | float
Children = Children_ | list[Children_]
SectionRender = Literal["accordion", "tabs", "steps"]
Position = Literal["top", "bottom", "none"]


class ModelForm(html.Div):
    """Create a Dash form from a pydantic model.

    Parameters
    ----------
    item: BaseModel | type[BaseModel]
        The model to create the form from, can be the model class or an instance of the class.
        If the class is passed, the form will be empty. If an instance is passed, the form will be pre-filled
        with existing values.
    aio_id: str
        All-in-one component ID
    form_id: str
        Form ID, can be used to create multiple forms on the same page. When working with databases
        this could be the document / record ID.
    fields_repr: dict[str, dict | BaseField] | None
        Mapping between field name and field representation. If not provided, default field
        representations will be used based on the field annotation.
        See `fields.get_default_repr`.
    sections: Sections | None
        List of form sections (optional). See `Sections`.
    submit_on_enter: bool
        Whether to submit the form on enter. Default False.
        Note: this may break the behaviour of some fields (e.g. editable table), use with caution.
    excluded_fields: list[str] | None
        List of field names to exclude from the form altogether, optional.
    container_kwargs: dict | None
        Additional kwargs to pass to the containing div.
    read_only: bool | None
        Whether the form should be read only.
        True/False set all the fields to read only or not. None keeps the field setting.
    debounce_inputs: int | None
        Debounce inputs in milliseconds. Only works with DMC components that can be debounced.
    """

    class ids:
        """Model form ids."""

        form = partial(form_base_id, "_pydf-form")
        main = partial(form_base_id, "_pydf-main")
        errors = partial(form_base_id, "_pydf-errors")
        accordion = partial(form_base_id, "_pydf-accordion")
        tabs = partial(form_base_id, "_pydf-tabs")
        steps = partial(form_base_id, "_pydf-steps")
        steps_save = partial(form_base_id, "_pydf-steps-save")
        steps_next = partial(form_base_id, "_pydf-steps-next")
        steps_previous = partial(form_base_id, "_pydf-steps-previous")
        steps_nsteps = partial(form_base_id, "_pydf-steps-nsteps")
        model_store = partial(form_base_id, "_pydf-model-store")
        form_specs_store = partial(form_base_id, "_pydf-form-specs-store")

    def __init__(  # noqa: PLR0912, PLR0913, PLR0915
        self,
        item: BaseModel | type[BaseModel],
        aio_id: str,
        form_id: str,
        path: str = "",
        fields_repr: dict[str, Union["BaseField", dict]] | None = None,
        sections: Sections | None = None,
        submit_on_enter: bool = False,
        discriminator: str | None = None,
        excluded_fields: list[str] | None = None,
        container_kwargs: dict | None = None,
        read_only: bool | None = None,
        debounce_inputs: int | None = None,
    ) -> None:
        with contextlib.suppress(Exception):
            if issubclass(item, BaseModel):
                item = item.model_construct()

        fields_repr = fields_repr or {}

        subitem_cls, disc_vals = self.get_discriminated_subitem_cls(item=item, path=path, discriminator=discriminator)
        field_inputs = self.render_fields(
            item=item,
            aio_id=aio_id,
            form_id=form_id,
            path=path,
            subitem_cls=subitem_cls,
            disc_vals=disc_vals,
            fields_repr=fields_repr,
            excluded_fields=excluded_fields,
            discriminator=discriminator,
            read_only=read_only,
            debounce_inputs=debounce_inputs,
        )

        if not sections or not any([f for f in field_inputs if f in s.fields] for s in sections.sections if s.fields):
            children = [self.grid(list(field_inputs.values()))]
        else:
            children = self.handle_sections(
                field_inputs=field_inputs, sections=sections, aio_id=aio_id, form_id=form_id, path=path
            )

        children.extend(
            self.get_meta_children(
                subitem_cls=subitem_cls,
                fields_repr=fields_repr,
                sections=sections,
                item=item,
                aio_id=aio_id,
                form_id=form_id,
                path=path,
            )
        )

        super().__init__(
            children=children,
            **(
                {
                    "id": self.ids.form(aio_id, form_id, path),
                    "data-submitonenter": submit_on_enter,
                    "style": {"containerType": "inline-size"},
                }
                if not path
                else {}
            ),
            **(container_kwargs or {}),
        )

    @staticmethod
    def get_discriminated_subitem_cls(
        *, item: BaseModel, path: str, discriminator: str
    ) -> tuple[type[BaseModel], tuple]:
        """Get the subitem of a model at a given parent, handling type unions."""
        subitem_cls = get_subitem_cls(item.__class__, path, item=item)

        # Handle type unions
        disc_vals = None
        discriminator_value = None
        if Type.classify(subitem_cls, discriminator) == Type.DISCRIMINATED_MODEL:
            subitem = get_subitem(item, path)
            discriminator_value = None if subitem is None else getattr(subitem, discriminator, None)
            subitem_cls, disc_vals = handle_discriminated(
                item.__class__, path, subitem_cls, discriminator, discriminator_value
            )

        return subitem_cls, disc_vals

    @staticmethod
    def render_fields(  # noqa: PLR0913
        *,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        path: str,
        subitem_cls: type[BaseModel],
        disc_vals: list[str],
        fields_repr: dict[str, dict | BaseField],
        excluded_fields: list[str],
        discriminator: str | None,
        read_only: bool | None,
        debounce_inputs: int | None,
    ) -> dict[str, Component]:
        """Render each field in the form."""
        from dash_pydantic_form.fields import get_default_repr

        field_inputs = {}
        for field_name, field_info in subitem_cls.model_fields.items():
            if field_name in (excluded_fields or []):
                continue
            more_kwargs = {}
            if read_only is not None:
                more_kwargs["read_only"] = read_only
            if debounce_inputs is not None:
                more_kwargs["debounce"] = debounce_inputs
            # If discriminating field, ensure all discriminator values are shown
            # Also add required metadata for discriminator callback
            if disc_vals and field_name == discriminator:
                field_info = deepcopy(field_info)  # noqa: PLW2901
                field_info.annotation = Literal[disc_vals]
                more_kwargs |= {"n_cols": 4, "field_id_meta": "discriminator"}
            if field_name in fields_repr:
                if isinstance(fields_repr[field_name], dict):
                    field_repr = get_default_repr(field_info, **fields_repr[field_name], **more_kwargs)
                else:
                    field_repr = fields_repr[field_name]
                    if more_kwargs:
                        field_repr = field_repr.__class__(**(field_repr.model_dump() | more_kwargs))
            else:
                field_repr = get_default_repr(field_info, **more_kwargs)

            field_inputs[field_name] = field_repr.render(
                item=item,
                aio_id=aio_id,
                form_id=form_id,
                field=field_name,
                parent=path,
                field_info=field_info,
            )

        return field_inputs

    @classmethod
    def handle_sections(  # noqa: PLR0913
        cls,
        *,
        field_inputs: dict[str, Component],
        sections: Sections,
        aio_id: str,
        form_id: str,
        path: str,
    ):
        """Handle the sections."""
        fields_not_in_sections = set(field_inputs) - set(
            itertools.chain.from_iterable(s.fields for s in sections.sections)
        )

        if sections.render == "accordion":
            render_function = cls.render_accordion_sections
        elif sections.render == "tabs":
            render_function = cls.render_tabs_sections
        elif sections.render == "steps":
            render_function = cls.render_steps_sections
        else:
            raise ValueError("Only 'accordion', 'tabs' and 'steps' are supported for `section_render_type`.")

        sections_render = render_function(
            aio_id=aio_id,
            form_id=form_id,
            path=path,
            field_inputs=field_inputs,
            sections=sections,
            **sections.render_kwargs,
        )

        if sections.remaining_fields_position == "none" or not fields_not_in_sections:
            children = sections_render
        elif sections.remaining_fields_position == "top":
            children = [
                cls.grid([v for k, v in field_inputs.items() if k in fields_not_in_sections], mb="sm")
            ] + sections_render
        else:
            children = sections_render + [
                cls.grid([v for k, v in field_inputs.items() if k in fields_not_in_sections], mb="sm")
            ]

        return children

    @classmethod
    def get_meta_children(  # noqa: PLR0913
        cls,
        *,
        subitem_cls: type[BaseModel],
        fields_repr: dict[str, dict | BaseField],
        sections: Sections | None,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        path: str,
    ):
        """Get 'meta' form children used for passing data to callbacks."""
        from dash_pydantic_form.fields import get_default_repr

        children = []
        if not path:
            children.append(dcc.Store(id=cls.ids.main(aio_id, form_id)))
            children.append(dcc.Store(id=cls.ids.errors(aio_id, form_id)))
            children.append(dcc.Store(data=str(item.__class__), id=cls.ids.model_store(aio_id, form_id)))

        fields_repr_dicts = (
            {
                field_name: (
                    get_default_repr(subitem_cls.model_fields[field_name], **field_repr)
                    if isinstance(field_repr, dict)
                    else field_repr
                ).to_dict()
                for field_name, field_repr in fields_repr.items()
                if field_name in subitem_cls.model_fields
            }
            if fields_repr
            else None
        )

        children.append(
            dcc.Store(
                data={
                    "sections": sections.model_dump(mode="json") if sections else None,
                    "fields_repr": fields_repr_dicts,
                },
                id=cls.ids.form_specs_store(aio_id, form_id, path),
            )
        )

        return children

    @classmethod
    def grid(cls, children: Children, **kwargs):
        """Create the responsive grid for a field."""
        return dmc.SimpleGrid(children, className="pydantic-form-grid " + kwargs.pop("className", ""), **kwargs)

    @classmethod
    def render_accordion_sections(  # noqa: PLR0913
        cls,
        *,
        aio_id: str,
        form_id: str,
        path: str,
        field_inputs: dict[str, Component],
        sections: Sections,
        **_kwargs,
    ):
        """Render the form sections as accordion."""
        return [
            dmc.Accordion(
                [
                    dmc.AccordionItem(
                        [
                            dmc.AccordionControl(
                                dmc.Text(
                                    ([DashIconify(icon=section.icon)] if section.icon else []) + [section.name],
                                    style={"display": "flex", "alignItems": "center", "gap": "0.5rem"},
                                    fw=600,
                                ),
                            ),
                            dmc.AccordionPanel(
                                cls.grid([field_inputs[field] for field in section.fields if field in field_inputs]),
                            ),
                        ],
                        value=section.name,
                    )
                    for section in sections.sections
                ],
                value=[section.name for section in sections.sections if section.default_open],
                styles={
                    "control": {"padding": "0.5rem"},
                    "label": {"padding": 0},
                    "item": {
                        "border": "1px solid color-mix(in srgb, var(--mantine-color-gray-light), transparent 40%)",
                        "background": "color-mix(in srgb, var(--mantine-color-gray-light), transparent 80%)",
                        "marginBottom": "0.5rem",
                        "borderRadius": "0.25rem",
                    },
                    "content": {
                        "display": "flex",
                        "flexDirection": "column",
                        "gap": "0.375rem",
                        "padding": "0.125rem 0.5rem 0.5rem",
                    },
                },
                id=cls.ids.accordion(aio_id, form_id, path),
                multiple=True,
            )
        ]

    @classmethod
    def render_tabs_sections(  # noqa: PLR0913
        cls,
        *,
        aio_id: str,
        form_id: str,
        path: str,
        field_inputs: dict[str, Component],
        sections: Sections,
        **_kwargs,
    ):
        """Render the form sections as tabs."""
        value = sections.sections[0].name
        for section in sections.sections:
            if section.default_open:
                value = section.name
                break

        return [
            dmc.Tabs(
                [
                    dmc.TabsList(
                        [
                            dmc.TabsTab(
                                dmc.Text(
                                    ([DashIconify(icon=section.icon)] if section.icon else []) + [section.name],
                                    style={"display": "flex", "alignItems": "center", "gap": "0.5rem"},
                                ),
                                value=section.name,
                            )
                            for section in sections.sections
                        ]
                    ),
                    *[
                        dmc.TabsPanel(
                            cls.grid([field_inputs[field] for field in section.fields if field in field_inputs]),
                            value=section.name,
                        )
                        for section in sections.sections
                    ],
                ],
                value=value,
                styles={
                    "panel": {"padding": "1rem 0.5rem 0"},
                },
                id=cls.ids.tabs(aio_id, form_id, path),
            )
        ]

    @classmethod
    def render_steps_sections(  # noqa: PLR0913
        cls,
        *,
        aio_id: str,
        form_id: str,
        path: str,
        field_inputs: dict[str, Component],
        sections: Sections,
        additional_steps: list = None,
        **kwargs,
    ):
        """Render the form sections as steps."""
        additional_steps = additional_steps or []
        stepper_styles = deep_merge(
            {
                "root": {"display": "flex", "gap": "1.5rem", "padding": "0.75rem 0 2rem"},
                "content": {"flex": 1, "padding": 0},
                "steps": {"minWidth": 180},
                "step": {"cursor": "pointer"},
                "stepBody": {"padding-top": "0.6875rem"},
                "stepCompletedIcon": {"&>svg": {"width": 12}},
            },
            kwargs.get("styles", {}),
        )
        stepper = dmc.Stepper(
            id=cls.ids.steps(aio_id, form_id, path),
            active=0,
            orientation="vertical",
            size="sm",
            styles=stepper_styles,
            children=[
                dmc.StepperStep(
                    label=section.name,
                    icon=DashIconify(icon=section.icon) if section.icon else None,
                    children=cls.grid([field_inputs[field] for field in section.fields if field in field_inputs]),
                )
                for section in sections.sections
            ]
            + additional_steps,
            **kwargs,
        )

        return [
            html.Div(
                [
                    stepper,
                    dmc.Group(
                        [
                            dmc.Button(
                                "Back",
                                id=cls.ids.steps_previous(aio_id, form_id, path),
                                disabled=True,
                                size="compact-md",
                                leftSection=DashIconify(icon="carbon:arrow-left", height=16),
                            ),
                            dmc.Button(
                                "Next",
                                id=cls.ids.steps_next(aio_id, form_id, path),
                                size="compact-md",
                                rightSection=DashIconify(icon="carbon:arrow-right", height=16),
                            ),
                        ],
                        style={
                            "position": "absolute",
                            "top": f"calc({78 * (len(sections.sections) + len(additional_steps))}px + 1rem)",
                        },
                    ),
                    dcc.Store(data=len(sections.sections), id=cls.ids.steps_nsteps(aio_id, form_id, path)),
                ],
                style={"position": "relative"},
            )
        ]


clientside_callback(
    ClientsideFunction(namespace="pydf", function_name="stepsPreviousNext"),
    Output(ModelForm.ids.steps(MATCH, MATCH, MATCH), "active"),
    Input(ModelForm.ids.steps_previous(MATCH, MATCH, MATCH), "n_clicks"),
    Input(ModelForm.ids.steps_next(MATCH, MATCH, MATCH), "n_clicks"),
    State(ModelForm.ids.steps(MATCH, MATCH, MATCH), "active"),
    State(ModelForm.ids.steps_nsteps(MATCH, MATCH, MATCH), "data"),
    prevent_inital_call=True,
)

clientside_callback(
    ClientsideFunction(namespace="pydf", function_name="stepsDisable"),
    Output(ModelForm.ids.steps_previous(MATCH, MATCH, MATCH), "disabled"),
    Output(ModelForm.ids.steps_next(MATCH, MATCH, MATCH), "disabled"),
    Input(ModelForm.ids.steps(MATCH, MATCH, MATCH), "active"),
    State(ModelForm.ids.steps_nsteps(MATCH, MATCH, MATCH), "data"),
)

clientside_callback(
    ClientsideFunction(namespace="pydf", function_name="stepsClickListener"),
    Output(ModelForm.ids.steps(MATCH, MATCH, MATCH), "id"),
    Input(ModelForm.ids.steps(MATCH, MATCH, MATCH), "id"),
)

clientside_callback(
    ClientsideFunction(namespace="pydf", function_name="getValues"),
    Output(ModelForm.ids.main(MATCH, MATCH), "data"),
    Input(common_ids.value_field(MATCH, MATCH, ALL, ALL, ALL), "value"),
    Input(common_ids.checked_field(MATCH, MATCH, ALL, ALL, ALL), "checked"),
    Input(fields.Dict.ids.item_key(MATCH, MATCH, ALL, ALL, ALL), "value"),
    Input(BaseField.ids.visibility_wrapper(MATCH, MATCH, ALL, ALL, ALL), "style"),
)

clientside_callback(
    ClientsideFunction(namespace="pydf", function_name="listenToSubmit"),
    Output(ModelForm.ids.form(MATCH, MATCH), "id"),
    Input(ModelForm.ids.form(MATCH, MATCH), "id"),
    State(ModelForm.ids.form(MATCH, MATCH), "data-submitonenter"),
)

clientside_callback(
    ClientsideFunction(namespace="pydf", function_name="displayErrors"),
    Output(common_ids.value_field(MATCH, MATCH, ALL, ALL, ALL), "error"),
    Input(ModelForm.ids.errors(MATCH, MATCH), "data"),
    State(common_ids.value_field(MATCH, MATCH, ALL, ALL, ALL), "id"),
)


@callback(
    Output(fields.Model.ids.form_wrapper(MATCH, MATCH, MATCH, MATCH), "children"),
    Input(common_ids.value_field(MATCH, MATCH, MATCH, MATCH, "discriminator"), "value"),
    State(ModelForm.ids.main(MATCH, MATCH), "data"),
    State(ModelForm.ids.model_store(MATCH, MATCH), "data"),
    State(ModelForm.ids.form_specs_store(MATCH, MATCH, MATCH), "data"),
    prevent_initial_call=True,
)
def update_discriminated(val, form_data: dict, model_name: str, form_specs: dict):
    """Update discriminated form."""
    path: str = get_fullpath(ctx.triggered_id["parent"], ctx.triggered_id["field"])
    parts = path.split(SEP)
    # Update the form data with the new value as it wouldn't have been updated yet
    pointer = form_data
    for i, part in enumerate(parts):
        if i == len(parts) - 1:
            pointer[part] = val
        pointer = pointer[int(part) if part.isdigit() else part]

    # Create an instance of the model sith the form data using model_construct_recursive
    # to build it out as much as possible without failing on validation
    model_cls = get_model_cls(model_name)
    item = model_construct_recursive(form_data, model_cls)

    # Retrieve fields-repr and sections from the stored data
    fields_repr: dict[str, BaseField] = {
        k: BaseField.from_dict(v) for k, v in (form_specs["fields_repr"] or {}).items()
    }
    sections = Sections(**form_specs["sections"]) if form_specs["sections"] else None

    return ModelForm(
        item=item,
        aio_id=ctx.triggered_id["aio_id"],
        form_id=ctx.triggered_id["form_id"],
        path=ctx.triggered_id["parent"],
        discriminator=ctx.triggered_id["field"],
        sections=sections,
        fields_repr=fields_repr,
    )



from functools import partial
from typing import Any

import dash_mantine_components as dmc
from dash import MATCH, ClientsideFunction, Input, Output, clientside_callback, dcc, html
from dash.development.base_component import Component
from pydantic import BaseModel
from pydantic.fields import FieldInfo

from dash_pydantic_form import ids as common_ids
from dash_pydantic_form.fields.base_fields import BaseField


class MarkdownField(BaseField):
    """Markdown field."""

    full_width = True

    class ids(BaseField.ids):
        """Model list field ids."""

        preview = partial(common_ids.field_dependent_id, "_pydf-markdown-field-preview")

    def _render(  # noqa: PLR0913
        self,
        *,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str = "",
        field_info: FieldInfo | None = None,
    ) -> Component:
        """Render the markdown field."""
        value = self.get_value(item, field, parent)
        if self.read_only:
            return self._render_read_only(value, field, field_info)

        return html.Div(
            [
                dmc.Textarea(
                    value=value,
                    label=self.get_title(field_info, field_name=field),
                    description=self.get_description(field_info),
                    id=common_ids.value_field(aio_id, form_id, field, parent=parent),
                    autosize=True,
                    minRows=3,
                    **self.input_kwargs,
                ),
                dcc.Markdown(
                    value,
                    id=self.ids.preview(aio_id, form_id, field, parent=parent),
                    className="pydf-markdown-preview",
                ),
            ],
            style={"display": "grid", "gap": "1rem", "gridTemplateColumns": "1fr 1fr"},
        )

    @staticmethod
    def _get_value_repr(value: Any, _field_info: FieldInfo):
        if value is None:
            return "-"
        return dcc.Markdown(value)


clientside_callback(
    ClientsideFunction(namespace="pydf", function_name="sync"),
    Output(MarkdownField.ids.preview(MATCH, MATCH, MATCH, parent=MATCH), "children"),
    Input(common_ids.value_field(MATCH, MATCH, MATCH, parent=MATCH), "value"),
)



from functools import partial

from dash.dependencies import _Wildcard


def form_dependent_id(component: str, aio_id: str | _Wildcard, form_id: str | _Wildcard) -> dict:
    """A component id to do callbacks at the document level (e.g. edit/delete)."""
    return {"component": component, "aio_id": aio_id, "form_id": form_id}


def field_dependent_id(  # noqa: PLR0913
    component: str,
    aio_id: str | _Wildcard,
    form_id: str | _Wildcard,
    field: str | _Wildcard,
    parent: str | _Wildcard = "",
    meta: str | _Wildcard = "",
) -> dict:
    """A component id to do callbacks at the field level (e.g. in the form)."""
    return {
        "component": component,
        "aio_id": aio_id,
        "form_id": form_id,
        "field": field,
        "parent": parent,
        "meta": meta,
    }


value_field = partial(field_dependent_id, "_pydf-value-field")
checked_field = partial(field_dependent_id, "_pydf-checked-field")



import json
from datetime import date, time

import dash_mantine_components as dmc
from dash import Dash, Input, Output
from pydantic import BaseModel, Field
from selenium.webdriver.common.by import By

from dash_pydantic_form import ModelForm, fields, ids
from tests.utils import (
    check_ids_exist,
    set_input,
    stringify_id,
)


def test_di0001_basic_dict(dash_duo):
    """Test a basic list."""

    class Basic(BaseModel):
        a: dict[str, int] = Field(title="Dict", default_factory=dict)

    app = Dash(__name__)
    aio_id = "aio"
    form_id = "form"
    app.layout = dmc.MantineProvider(
        [
            ModelForm(Basic, aio_id=aio_id, form_id=form_id),
            dmc.Text(id="output"),
        ],
    )
    check_ids_exist(
        app.layout,
        [fields.List.ids.add(aio_id, form_id, "a")],
    )

    @app.callback(
        Output("output", "children"),
        Input(ModelForm.ids.main(aio_id, form_id), "data"),
    )
    def display(form_data):
        return json.dumps(form_data)

    dash_duo.start_server(app)

    dash_duo.wait_for_text_to_equal("#output", "{}")
    add_btn = dash_duo.driver.find_element(By.ID, stringify_id(fields.List.ids.add(aio_id, form_id, "a")))
    add_btn.click()
    set_input(dash_duo, ids.value_field(aio_id, form_id, "0", parent="a"), 123)
    set_input(dash_duo, fields.Dict.ids.item_key(aio_id, form_id, "a", meta="0"), "aa")
    dash_duo.wait_for_text_to_equal("#output", '{"a": {"aa": 123}}')
    add_btn.click()
    set_input(dash_duo, fields.Dict.ids.item_key(aio_id, form_id, "a", meta="1"), "bb")
    set_input(dash_duo, ids.value_field(aio_id, form_id, "1", parent="a"), 456)
    dash_duo.wait_for_text_to_equal("#output", '{"a": {"aa": 123, "bb": 456}}')

    del_btn0 = dash_duo.driver.find_element(By.ID, stringify_id(fields.List.ids.delete(aio_id, form_id, "a", meta="0")))
    # 2 clicks required, on to bring to the foreground one to actually click
    del_btn0.click()
    del_btn0.click()
    dash_duo.wait_for_text_to_equal("#output", '{"a": {"bb": 456}}')


def test_di0002_dict_all_scalar_types(dash_duo):
    """Test a list of all scalar types."""

    class Basic(BaseModel):
        a: dict[str, int]
        b: dict[str, str]
        c: dict[str, date]
        d: dict[str, time]

    app = Dash(__name__)
    item = Basic(
        a={"int0": 1, "int1": 2},
        b={"str0": "a", "str1": "b"},
        c={"date0": date(2020, 1, 1), "date1": date(2020, 1, 2)},
        d={"time0": "12:00", "time1": "13:00"},
    )
    aio_id = "aio"
    form_id = "form"
    app.layout = dmc.MantineProvider(
        [
            ModelForm(item, aio_id=aio_id, form_id=form_id),
            dmc.Text(id="output"),
        ],
    )

    @app.callback(
        Output("output", "children"),
        Input(ModelForm.ids.main(aio_id, form_id), "data"),
    )
    def display(form_data):
        return json.dumps(form_data)

    dash_duo.start_server(app)

    dash_duo.wait_for_text_to_equal(
        "#output",
        '{"a": {"int0": 1, "int1": 2}, "b": {"str0": "a", "str1": "b"}, '
        '"c": {"date0": "2020-01-01", "date1": "2020-01-02"}, '
        '"d": {"time0": "2000-01-01T12:00:00", "time1": "2000-01-01T13:00:00"}}',
    )


def test_di0003_model_dict(dash_duo):
    """Test a model list."""

    class Nested(BaseModel):
        b: dict[str, str] = Field(default_factory=dict)

    class Basic(BaseModel):
        a: dict[str, Nested] = Field(default_factory=dict)

    app = Dash(__name__)
    item = Basic()
    aio_id = "aio"
    form_id = "form"
    app.layout = dmc.MantineProvider(
        [
            ModelForm(item, aio_id=aio_id, form_id=form_id),
            dmc.Text(id="output"),
        ],
    )

    @app.callback(
        Output("output", "children"),
        Input(ModelForm.ids.main(aio_id, form_id), "data"),
    )
    def display(form_data):
        return json.dumps(form_data, sort_keys=True)

    dash_duo.start_server(app)

    dash_duo.wait_for_text_to_equal("#output", "{}")

    dash_duo.driver.find_element(By.ID, stringify_id(fields.List.ids.add(aio_id, form_id, "a"))).click()
    dash_duo.driver.find_element(By.CSS_SELECTOR, ".mantine-Accordion-control").click()
    dash_duo.driver.find_element(By.ID, stringify_id(fields.List.ids.add(aio_id, form_id, "b", parent="a:0"))).click()
    set_input(dash_duo, ids.value_field(aio_id, form_id, "0", parent="a:0:b"), 1)
    set_input(dash_duo, fields.Dict.ids.item_key(aio_id, form_id, "a", meta="0"), "a0")
    set_input(dash_duo, fields.Dict.ids.item_key(aio_id, form_id, "b", parent="a:0", meta="0"), "b0")
    dash_duo.wait_for_text_to_equal("#output", '{"a": {"a0": {"b": {"b0": "1"}}}}')



from typing import Literal

import dash_mantine_components as dmc
from dash import Dash
from pydantic import BaseModel, Field

from dash_pydantic_form import ModelForm, ids
from tests.utils import (
    check_ids_absent,
    check_ids_exist,
    set_checkbox,
    set_input,
    set_select,
)


def test_du0001_basic_discriminated_union(dash_duo):
    """Test a basic discriminated union."""

    class Cat(BaseModel):
        species: Literal["cat"]
        meows: bool = True

    class Dog(BaseModel):
        species: Literal["dog"]
        barks: bool = True

    class Basic(BaseModel):
        pet: Cat | Dog | None = Field(title="Pet", discriminator="species", default=None)

    app = Dash(__name__)
    item = Basic()
    aio_id = "basic"
    form_id = "form"
    app.layout = dmc.MantineProvider(ModelForm(item, aio_id=aio_id, form_id=form_id))
    check_ids_absent(
        app.layout,
        [
            ids.checked_field(aio_id, form_id, "meows", parent="pet"),
            ids.checked_field(aio_id, form_id, "barks", parent="pet"),
        ],
    )

    dash_duo.start_server(app)
    set_select(dash_duo, ids.value_field(aio_id, form_id, "species", parent="pet", meta="discriminator"), "cat")
    set_checkbox(dash_duo, ids.checked_field(aio_id, form_id, "meows", parent="pet"), False)
    set_select(dash_duo, ids.value_field(aio_id, form_id, "species", parent="pet", meta="discriminator"), "dog")
    set_checkbox(dash_duo, ids.checked_field(aio_id, form_id, "barks", parent="pet"), False)


def test_du0002_discriminated_union_nested(dash_duo):
    """Test a discriminated union nested in a model."""

    class Cat(BaseModel):
        species: Literal["cat"]
        meows: bool = True

    class Dog(BaseModel):
        species: Literal["dog"]
        barks: bool = True

    class Nested(BaseModel):
        pet: Cat | Dog | None = Field(title="Pet", discriminator="species", default=None)

    class Basic(BaseModel):
        nested: Nested

    app = Dash(__name__)
    item = Basic(nested={"pet": {"species": "cat"}})
    aio_id = "basic"
    form_id = "form"
    app.layout = dmc.MantineProvider(ModelForm(item, aio_id=aio_id, form_id=form_id))
    check_ids_exist(
        app.layout,
        [ids.checked_field(aio_id, form_id, "meows", parent="nested:pet")],
    )
    dash_duo.start_server(app)
    set_checkbox(dash_duo, ids.checked_field(aio_id, form_id, "meows", parent="nested:pet"), False)
    set_select(dash_duo, ids.value_field(aio_id, form_id, "species", parent="nested:pet", meta="discriminator"), "dog")
    set_checkbox(dash_duo, ids.checked_field(aio_id, form_id, "barks", parent="nested:pet"), False)


def test_du0003_discriminated_union_with_nested_submodel(dash_duo):
    """Test a discriminated union with a nested submodel."""

    class CatOptions(BaseModel):
        meows: bool = True

    class DogOptions(BaseModel):
        barks: bool = True

    class Cat(BaseModel):
        species: Literal["cat"]
        options: CatOptions = Field(default_factory=CatOptions)

    class Dog(BaseModel):
        species: Literal["dog"]
        options: DogOptions = Field(default_factory=DogOptions)

    class Basic(BaseModel):
        pet: Cat | Dog | None = Field(title="Pet", discriminator="species", default=None)

    app = Dash(__name__)
    item = Basic(pet={"species": "cat"})
    aio_id = "basic"
    form_id = "form"
    app.layout = dmc.MantineProvider(ModelForm(item, aio_id=aio_id, form_id=form_id))
    check_ids_exist(
        app.layout,
        [ids.checked_field(aio_id, form_id, "meows", parent="pet:options")],
    )

    dash_duo.start_server(app)
    set_checkbox(dash_duo, ids.checked_field(aio_id, form_id, "meows", parent="pet:options"), False)
    set_select(dash_duo, ids.value_field(aio_id, form_id, "species", parent="pet", meta="discriminator"), "dog")
    set_checkbox(dash_duo, ids.checked_field(aio_id, form_id, "barks", parent="pet:options"), False)


def test_du004_discriminated_in_discriminated(dash_duo):
    """Test a model with nesting of discriminated unions."""

    class BlackOptions(BaseModel):
        color: Literal["black"]
        a: str = "A"

    class OrangeOptions(BaseModel):
        color: Literal["orange"]
        b: str = "B"

    class Cat(BaseModel):
        species: Literal["cat"]
        options: BlackOptions | OrangeOptions | None = Field(discriminator="color", default=None)

    class Dog(BaseModel):
        species: Literal["dog"]
        barks: bool = True

    class Basic(BaseModel):
        pet: Cat | Dog | None = Field(title="Pet", discriminator="species", default=None)

    app = Dash(__name__)
    item = Basic(pet={"species": "cat", "options": {"color": "black"}})
    aio_id = "basic"
    form_id = "form"
    app.layout = dmc.MantineProvider(ModelForm(item, aio_id=aio_id, form_id=form_id))
    check_ids_exist(
        app.layout,
        [
            ids.value_field(aio_id, form_id, "a", parent="pet:options"),
            ids.value_field(aio_id, form_id, "color", parent="pet:options", meta="discriminator"),
            ids.value_field(aio_id, form_id, "species", parent="pet", meta="discriminator"),
        ],
    )

    dash_duo.start_server(app)
    set_input(dash_duo, ids.value_field(aio_id, form_id, "a", parent="pet:options"), "123")
    set_select(
        dash_duo, ids.value_field(aio_id, form_id, "color", parent="pet:options", meta="discriminator"), "orange"
    )
    set_input(dash_duo, ids.value_field(aio_id, form_id, "b", parent="pet:options"), "456")
    set_select(dash_duo, ids.value_field(aio_id, form_id, "species", parent="pet", meta="discriminator"), "dog")
    set_checkbox(dash_duo, ids.checked_field(aio_id, form_id, "barks", parent="pet"), False)



import logging
import re
from copy import deepcopy
from datetime import date, time
from enum import Enum
from numbers import Number
from types import UnionType
from typing import Any, Literal, Union, get_args, get_origin

from pydantic import BaseModel, ValidationError, create_model
from pydantic_core import PydanticUndefined

SEP = ":"


class Type(Enum):
    """Types of fields."""

    SCALAR = "scalar"
    LITERAL = "literal"
    MODEL = "model"
    UNKNOWN = "unknown"
    DISCRIMINATED_MODEL = "discriminated_model"
    MODEL_LIST = "model_list"
    SCALAR_LIST = "scalar_list"
    LITERAL_LIST = "literal_list"
    UNKOWN_LIST = "unknown_list"
    MODEL_DICT = "model_dict"
    SCALAR_DICT = "scalar_dict"
    LITERAL_DICT = "literal_dict"
    UNKOWN_DICT = "unkown_dict"

    @classmethod
    def classify(cls, annotation: type, discriminator: str | None = None, depth: int = 0) -> bool:  # noqa: PLR0911, PLR0912
        """Classify a value as a field type."""
        annotation = get_non_null_annotation(annotation)

        if is_subclass(annotation, str | Number | bool | date | time):
            return cls.SCALAR

        if (get_origin(annotation) == Literal) | is_subclass(annotation, Enum):
            return cls.LITERAL

        if is_subclass(annotation, BaseModel):
            return cls.MODEL

        if get_origin(annotation) in [Union, UnionType]:
            if discriminator and all(is_subclass(x, BaseModel) for x in get_args(annotation)):
                return cls.DISCRIMINATED_MODEL
            if all(is_subclass(x, str | Number) for x in get_args(annotation)):
                return cls.SCALAR

        if get_origin(annotation) is list and not depth:
            ann_args = get_args(annotation)
            if not ann_args:
                return cls.UNKOWN_LIST
            args_type = Type.classify(ann_args[0], depth=1)
            if args_type == Type.SCALAR:
                return cls.SCALAR_LIST
            if args_type == Type.LITERAL:
                return cls.LITERAL_LIST
            if args_type == Type.MODEL:
                return cls.MODEL_LIST
            return cls.UNKOWN_LIST

        if get_origin(annotation) is dict and not depth:
            ann_args = get_args(annotation)
            if not ann_args:
                return cls.UNKOWN_DICT
            args_type = Type.classify(ann_args[1], depth=1)
            if args_type == Type.SCALAR:
                return cls.SCALAR_DICT
            if args_type == Type.LITERAL:
                return cls.LITERAL_DICT
            if args_type == Type.MODEL:
                return cls.MODEL_DICT
            return cls.UNKOWN_DICT

        return cls.UNKNOWN


def deep_merge(dict1: dict, dict2: dict) -> dict:
    """Deep merge two dictionaries, the second input is given priority."""
    dict_final = deepcopy(dict1)
    for key, val in dict2.items():
        if isinstance(val, dict):
            dict_final[key] = deep_merge(dict_final.get(key, {}), val)
        else:
            dict_final[key] = val
    return dict_final


def deep_diff(dict1: dict, dict2: dict) -> dict[str, dict | tuple[Any, Any]]:
    """Compute the deep difference between two dictionaries."""
    diff = {}
    for key in dict1.keys() | dict2.keys():
        if isinstance(dict1.get(key), dict) and isinstance(dict2.get(key), dict):
            sub_diff = deep_diff(dict1[key], dict2[key])
            if sub_diff:
                diff[key] = sub_diff
        elif dict1.get(key) != dict2.get(key):
            diff[key] = (dict1.get(key), dict2.get(key))
    return diff


def get_non_null_annotation(annotation: type[Any]) -> type[Any]:
    """Get a non-null annotation.

    e.g., get_non_null_annotation(Optional[str]) = str
    """
    if get_origin(annotation) in [Union, UnionType]:
        args = tuple(x for x in get_args(annotation) if x is not type(None))
        if len(args) == 1:
            return args[0]
        return Union[args]  # noqa: UP007
    return annotation


def get_model_value(item: BaseModel, field: str, parent: str, allow_default: bool = True):  # noqa: PLR0911
    """Get the value of a model (parent, field) pair.

    Parameters
    ----------
    item: BaseModel
        The object to get the value from
    field: str
        The field name
    parent: str
        The parent of the field (for nested fields), in dot notation
    allow_default: bool
        Allow to return the default value, when the object has been created with model_construct.
    """
    try:
        subitem = get_subitem(item, parent)
        if isinstance(subitem, BaseModel):
            return subitem[field]
        if isinstance(subitem, dict) and isinstance(field, int):
            return list(subitem.values())[field]
        if isinstance(subitem, list) and isinstance(field, int):
            return subitem[field]
        return subitem[field]
    except:
        if allow_default:
            subitem_cls = get_subitem_cls(item.__class__, parent, item=item)
            if not is_subclass(subitem_cls, BaseModel):
                return None
            field_info = subitem_cls.model_fields[field]
            if field_info.default is not PydanticUndefined:
                return field_info.default
            if field_info.default_factory:
                return field_info.default_factory()
            return None
        raise


def get_subitem(item: BaseModel | list | dict, parent: str) -> BaseModel:
    """Get the subitem of a model at a given parent.

    e.g., get_subitem(person, "au_metadata") = AUMetadata(param1=True, param2=False)
    """
    if parent == "":
        return item

    path = parent.split(SEP)

    first_part = path[0]
    if isinstance(first_part, str) and first_part.isdigit():
        first_part = int(first_part)

    if isinstance(item, BaseModel):
        next_item = getattr(item, first_part)
    elif isinstance(item, dict) and isinstance(first_part, int):
        next_item = list(item.values())[first_part]
    elif isinstance(item, list) and isinstance(first_part, int):
        next_item = item[first_part]
    else:
        next_item = item.get(first_part)

    if len(path) == 1 or next_item is None:
        return next_item

    return get_subitem(next_item, SEP.join(path[1:]))


def is_idx_template(val: str):
    """Check if a string is an index template."""
    return bool(re.findall(r"^\{\{[\w|\{\}]+\}\}$", val))


def get_subitem_cls(model: type[BaseModel], parent: str, item: BaseModel | None = None) -> type[BaseModel]:
    """Get the subitem class of a model at a given parent.

    e.g., get_subitem_cls(Person, "au_metadata") = AUMetadata
    """
    if parent == "":
        return model

    path = parent.split(SEP)

    first_part = path[0]
    second_part = None

    if len(path) == 1:
        ann = get_non_null_annotation(model.model_fields[first_part].annotation)
        return ann

    second_part = path[1]
    if isinstance(second_part, str) and second_part.isdigit():
        second_part = int(second_part)

    field_info = model.model_fields[first_part]
    first_annotation = get_non_null_annotation(field_info.annotation)
    try:
        subitem = get_subitem(item, first_part) if item is not None else None
    except:  # noqa: E722
        subitem = None
    if Type.classify(first_annotation, field_info.discriminator) == Type.DISCRIMINATED_MODEL:
        if not item:
            raise TypeError("Discriminated models with nesting need passing item data to be displayed")
        discriminator_value = None if subitem is None else getattr(subitem, field_info.discriminator, None)
        subitem_cls, _ = handle_discriminated(
            item.__class__, path, first_annotation, field_info.discriminator, discriminator_value
        )
        return get_subitem_cls(subitem_cls, SEP.join(path[1:]), item=subitem)
    if (
        get_origin(first_annotation) is list
        and (isinstance(second_part, int) or is_idx_template(second_part))
        and get_args(first_annotation)
    ):
        return get_subitem_cls(
            get_non_null_annotation(get_args(first_annotation)[0]),
            SEP.join(path[2:]),
            item=subitem,
        )
    if get_origin(first_annotation) is dict and (isinstance(second_part, int | str)) and get_args(first_annotation):
        return get_subitem_cls(
            get_non_null_annotation(get_args(first_annotation)[1]),
            SEP.join(path[2:]),
            item=subitem,
        )
    return get_subitem_cls(first_annotation, SEP.join(path[1:]), item=subitem)


def handle_discriminated(model: type[BaseModel], parent: str, annotation: type, disc_field: str, disc_val: Any):
    """Handle a discriminated model."""
    all_vals = set()
    out = None
    for possible in get_args(annotation):
        if not get_origin(possible.model_fields[disc_field].annotation) == Literal:
            raise ValueError("Discriminator must be a Literal")

        vals = get_args(possible.model_fields[disc_field].annotation)
        all_vals = all_vals.union(vals)
        if disc_val is not None and disc_val in vals:
            out = possible

    all_vals = tuple(all_vals)

    if disc_val is None:
        return create_model(
            f"{model.__name__}{parent.replace(SEP, ' ').title().replace(' ', '')}Discriminator",
            **{disc_field: (Literal[tuple(all_vals)], ...)},
        ), all_vals

    if out is None:
        raise ValueError(f"Invalid discriminator value: {disc_val}")

    return out, all_vals


def get_fullpath(*parts, sep: str = SEP):
    """Creates the full path of a field from its name and parent."""
    return sep.join([str(p) for p in parts]).strip(sep)


def get_all_subclasses(cls: type):
    """Get all subclasses of a class."""
    all_subclasses = []

    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))

    return all_subclasses


def get_model_cls(str_repr: str) -> type[BaseModel]:
    """Get the model class from a string representation."""
    return next(cls for cls in get_all_subclasses(BaseModel) if str(cls) == str_repr)


def is_subclass(cls: type, base_cls: type) -> bool:
    """Check if a class is a subclass of another class, handling issubclass errors."""
    try:
        return issubclass(cls, base_cls)
    except TypeError:
        return False


def model_construct_recursive(data: dict, data_model: type[BaseModel]):
    """Construct a model recursively."""
    updated = deepcopy(data)
    for key, val in data.items():
        if key not in data_model.model_fields:
            continue

        field_info = data_model.model_fields[key]
        ann = get_non_null_annotation(field_info.annotation)
        type_ = Type.classify(ann, field_info.discriminator)
        if type_ == Type.MODEL:
            updated[key] = model_construct_recursive(val, ann)
        elif type_ == Type.DISCRIMINATED_MODEL and field_info.discriminator in val:
            disc_val = val[field_info.discriminator]
            out = None
            for possible in get_args(ann):
                if not get_origin(possible.model_fields[field_info.discriminator].annotation) == Literal:
                    raise ValueError("Discriminator must be a Literal")

                if disc_val in get_args(possible.model_fields[field_info.discriminator].annotation):
                    out = possible
                    break
            if out is not None:
                updated[key] = model_construct_recursive(val, possible)
        elif type_ == Type.MODEL_LIST and isinstance(val, list):
            updated[key] = [model_construct_recursive(vv, get_args(ann)[0]) for vv in val]

    return data_model.model_construct(**updated)


def from_form_data(data: dict, data_model: type[BaseModel]):
    """Construct a model from form data, allowing to use default values when validation of a field fails."""
    try:
        return data_model.model_validate(data)
    except ValidationError as exc:
        data_with_defaults = deepcopy(data)
        defaulted_fields = []
        for error in exc.errors():
            path_parts = [str(x) for x in error["loc"]]
            for i in range(len(path_parts), 0, -1):
                parts_extract = path_parts[:i]
                path = SEP.join(parts_extract)
                try:
                    failing_model = get_subitem_cls(data_model, SEP.join(parts_extract[:-1]))
                except AttributeError:
                    continue
                if not is_subclass(failing_model, BaseModel):
                    continue
                field = failing_model.model_fields[parts_extract[-1]]
                if field.default != PydanticUndefined:
                    set_at_path(data_with_defaults, path, field.default)
                    defaulted_fields.append(path)
                    break
                if field.default_factory is not None:
                    set_at_path(data_with_defaults, path, field.default_factory())
                    defaulted_fields.append(path)
                    break

        if defaulted_fields:
            logging.info(
                "Could not validate the following fields: %s for %s, using default values instead.",
                defaulted_fields,
                data_model.__name__,
            )
        return data_model.model_validate(data_with_defaults)


def set_at_path(data: dict, path: str, value: Any):
    """Set a value at a path in a dictionary."""
    parts = path.split(SEP)
    pointer = data
    for part in parts[:-1]:
        pointer = pointer.get(part, pointer[int(part)])
    pointer[parts[-1]] = value



import uuid
from collections.abc import Callable
from functools import partial
from typing import Literal

import dash_mantine_components as dmc
from dash import (
    ALL,
    MATCH,
    ClientsideFunction,
    Input,
    Output,
    State,
    clientside_callback,
    dcc,
    html,
)
from dash.development.base_component import Component
from dash_iconify import DashIconify
from plotly.io.json import to_json_plotly
from pydantic import BaseModel, Field
from pydantic.fields import FieldInfo

from dash_pydantic_form import ids as common_ids
from dash_pydantic_form.fields.base_fields import BaseField
from dash_pydantic_form.form_section import Sections
from dash_pydantic_form.utils import (
    Type,
    get_fullpath,
    get_subitem_cls,
)


class ListField(BaseField):
    """List field, used for list of nested models or scalars.

    Optional attributes:
    * render_type (one of 'accordion', 'modal', 'list', default 'accordion')
        new render types can be defined by extending this class and overriding
        the following methods: _contents_renderer and render_type_item_mapper
    * fields_repr, mapping between field name and field representation
    * sections, list of FormSection for the NestedModel form
    * items_deletable, whether the items can be deleted (bool, default True)
    * items_creatable, whether new items can be created (bool, default True)
    """

    render_type: Literal["accordion", "modal", "list", "scalar"] = Field(
        default="accordion",
        description=(
            "How to render the list of items. One  of 'accordion', 'modal', 'list' for a list of models. "
            "Should be set to 'scalar' for a list of scalars."
        ),
    )
    fields_repr: dict[str, dict | BaseField] | None = Field(
        default=None,
        description="Fields representation, mapping between field name and field representation for the nested fields.",
    )
    sections: Sections | None = Field(default=None, description="Sub-form sections.")
    items_deletable: bool = Field(default=True, description="Whether the items can be deleted.")
    items_creatable: bool = Field(default=True, description="Whether new items can be created.")

    full_width = True

    def model_post_init(self, _context):
        """Model post init."""
        super().model_post_init(_context)
        if self.fields_repr is None:
            self.fields_repr = {}
        if self.read_only:
            self.items_deletable = False
            self.items_creatable = False

    class ids(BaseField.ids):
        """Model list field ids."""

        wrapper = partial(common_ids.field_dependent_id, "_pydf-list-field-wrapper")
        delete = partial(common_ids.field_dependent_id, "_pydf-list-field-delete")
        edit = partial(common_ids.field_dependent_id, "_pydf-list-field-edit")
        modal = partial(common_ids.field_dependent_id, "_pydf-list-field-modal")
        accordion_parent_text = partial(common_ids.field_dependent_id, "_pydf-list-field-accordion-text")
        modal_parent_text = partial(common_ids.field_dependent_id, "_pydf-list-field-modal-text")
        modal_save = partial(common_ids.field_dependent_id, "_pydf-list-field-modal-save")
        add = partial(common_ids.field_dependent_id, "_pydf-list-field-add")
        template_store = partial(common_ids.field_dependent_id, "_pydf-list-field-template-store")

    @classmethod
    def accordion_item(  # noqa: PLR0913
        cls,
        *,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str,
        index: int,
        value: BaseModel,
        fields_repr: dict[str, dict | BaseField] | None = None,
        sections: Sections | None = None,
        items_deletable: bool = True,
        read_only: bool | None = None,
        **_kwargs,
    ):
        """Create an accordion item for the model list field."""
        from dash_pydantic_form import ModelForm

        new_parent = get_fullpath(parent, field, index)
        return dmc.AccordionItem(
            # Give a random unique value to the item, prepended by uuid: so that the callback
            # to add new items works
            value="uuid:" + uuid.uuid4().hex,
            style={"position": "relative"},
            className="pydf-model-list-accordion-item",
            children=[
                dmc.AccordionControl(
                    [dmc.Text(str(value), id=cls.ids.accordion_parent_text(aio_id, form_id, "", parent=new_parent))]
                    + items_deletable
                    * [
                        dmc.ActionIcon(
                            DashIconify(icon="carbon:trash-can", height=16),
                            color="red",
                            style={
                                "position": "absolute",
                                "top": "50%",
                                "transform": "translateY(-50%)",
                                "right": "2.5rem",
                            },
                            variant="light",
                            size="sm",
                            id=cls.ids.delete(aio_id, form_id, field, parent=parent, meta=index),
                            className="pydf-model-list-accordion-item-delete",
                        ),
                    ],
                    pos="relative",
                ),
                dmc.AccordionPanel(
                    ModelForm(
                        item=item,
                        aio_id=aio_id,
                        form_id=form_id,
                        path=new_parent,
                        fields_repr=fields_repr,
                        sections=sections,
                        read_only=read_only,
                    ),
                ),
            ],
        )

    @classmethod
    def list_item(  # noqa: PLR0913
        cls,
        *,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str,
        index: int,
        fields_repr: dict[str, dict | BaseField] | None = None,
        sections: Sections | None = None,
        items_deletable: bool = True,
        read_only: bool | None = None,
        **_kwargs,
    ):
        """Create an item with bare forms for the model list field."""
        from dash_pydantic_form import ModelForm

        new_parent = get_fullpath(parent, field, index)
        return dmc.Group(
            [
                ModelForm(
                    item=item,
                    aio_id=aio_id,
                    form_id=form_id,
                    path=new_parent,
                    fields_repr=fields_repr,
                    sections=sections,
                    container_kwargs={"style": {"flex": 1}},
                    read_only=read_only,
                ),
            ]
            + items_deletable
            * [
                dmc.ActionIcon(
                    DashIconify(icon="carbon:trash-can", height=16),
                    color="red",
                    variant="light",
                    size="sm",
                    id=cls.ids.delete(aio_id, form_id, field, parent=parent, meta=index),
                ),
            ],
            gap="sm",
            align="top",
            className="pydf-model-list-list-item",
        )

    @classmethod
    def modal_item(  # noqa: PLR0913
        cls,
        *,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str,
        index: int,
        value: BaseModel,
        opened: bool = False,
        fields_repr: dict[str, dict | BaseField] | None = None,
        sections: Sections | None = None,
        items_deletable: bool = True,
        read_only: bool | None = None,
        **_kwargs,
    ):
        """Create an item with bare forms for the model list field."""
        from dash_pydantic_form import ModelForm

        new_parent = get_fullpath(parent, field, index)
        return dmc.Paper(
            dmc.Group(
                [
                    dmc.Text(
                        str(value),
                        style={
                            "flex": 1,
                            "overflow": "hidden",
                            "textOverflow": "ellipsis",
                            "whiteSpace": "nowrap",
                        },
                        id=cls.ids.modal_parent_text(aio_id, form_id, "", parent=new_parent),
                    ),
                    dmc.Group(
                        [
                            dmc.ActionIcon(
                                DashIconify(icon="carbon:view" if read_only else "carbon:edit", height=16),
                                variant="light",
                                size="sm",
                                id=cls.ids.edit(aio_id, form_id, "", parent=new_parent),
                                className="pydf-model-list-modal-item-btn",
                            ),
                        ]
                        + items_deletable
                        * [
                            dmc.ActionIcon(
                                DashIconify(icon="carbon:trash-can", height=16),
                                color="red",
                                variant="light",
                                size="sm",
                                id=cls.ids.delete(aio_id, form_id, field, parent=parent, meta=index),
                                className="pydf-model-list-modal-item-btn",
                            ),
                        ],
                        gap="0.5rem",
                    ),
                    dmc.Modal(
                        [
                            ModelForm(
                                item=item,
                                aio_id=aio_id,
                                form_id=form_id,
                                path=new_parent,
                                fields_repr=fields_repr,
                                sections=sections,
                                read_only=read_only,
                            ),
                            dmc.Group(
                                dmc.Button(
                                    "Save",
                                    leftSection=DashIconify(icon="carbon:save"),
                                    id=cls.ids.modal_save(aio_id, form_id, "", parent=new_parent),
                                ),
                                justify="right",
                                mt="sm",
                            ),
                        ],
                        title=str(value),
                        id=cls.ids.modal(aio_id, form_id, "", parent=new_parent),
                        style={"--modal-size": "min(calc(100vw - 4rem), 1150px)"},
                        opened=opened,
                    ),
                ],
                gap="sm",
                align="top",
            ),
            withBorder=True,
            radius="sm",
            p="xs",
            className="pydf-model-list-modal-item",
        )

    @classmethod
    def scalar_item(  # noqa: PLR0913
        cls,
        *,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str,
        index: int,
        items_deletable: bool = True,
        read_only: bool | None = None,
        input_kwargs: dict,
        **kwargs,
    ):
        """Create an item for a scalar list."""
        from dash_pydantic_form.fields import get_default_repr

        scalar_cls = get_subitem_cls(item.__class__, get_fullpath(parent, field, index), item=item)
        field_repr = get_default_repr(
            None, annotation=scalar_cls, read_only=read_only, title="", input_kwargs=input_kwargs, **kwargs
        )
        child = field_repr.render(
            item=item,
            aio_id=aio_id,
            form_id=form_id,
            field=index,
            parent=get_fullpath(parent, field),
            field_info=FieldInfo.from_annotation(scalar_cls),
        )
        if not child.style:
            child.style = {}
        child.style |= {"flex": "1 1 60%"}
        return dmc.Group(
            [child]
            + items_deletable
            * [
                dmc.ActionIcon(
                    DashIconify(icon="carbon:close", height=16),
                    color="red",
                    variant="light",
                    size="sm",
                    mt="0.375rem",
                    id=cls.ids.delete(aio_id, form_id, field, parent=parent, meta=index),
                    className="pydf-model-list-scalar-item-delete",
                ),
            ],
            gap=0,
            align="top",
            className="pydf-model-list-scalar-item",
            wrap="none",
        )

    def _contents_renderer(self, renderer_type: str) -> Callable:
        """Create a renderer for the model list field."""
        raise NotImplementedError(
            "Only the default renderers (accordion, list, modals) are implemented. "
            "Override this method to create custom renderers."
        )

    @classmethod
    def render_type_item_mapper(cls, render_type: str) -> dict[str, Callable]:
        """Mapping between render type and renderer function."""
        return getattr(cls, f"{render_type}_item")

    def _render(  # noqa: PLR0913
        self,
        *,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str = "",
        field_info: FieldInfo,
    ) -> Component:
        """Create a form field of type checklist to interact with the model field."""
        type_ = Type.classify(field_info.annotation, field_info.discriminator)
        if type_ == Type.MODEL_LIST and self.render_type == "scalar":
            raise ValueError("Cannot render model list as scalar")
        if type_ != Type.MODEL_LIST and self.render_type != "scalar":
            raise ValueError("Cannot render non model list as non scalar")

        value: list = self.get_value(item, field, parent) or []

        class_name = "pydf-model-list-wrapper" + (" required" if self.is_required(field_info) else "")
        if self.render_type == "accordion":
            contents = dmc.Accordion(
                [
                    self.accordion_item(
                        item=item,
                        aio_id=aio_id,
                        form_id=form_id,
                        field=field,
                        parent=parent,
                        index=i,
                        value=val,
                        fields_repr=self.fields_repr,
                        sections=self.sections,
                        items_deletable=self.items_deletable,
                        read_only=self.read_only,
                    )
                    for i, val in enumerate(value)
                ],
                id=self.ids.wrapper(aio_id, form_id, field, parent=parent),
                value=None,
                styles={
                    "control": {"padding": "0.5rem"},
                    "label": {"padding": 0},
                    "item": {
                        "border": "1px solid color-mix(in srgb, var(--mantine-color-gray-light), transparent 40%)",
                        "background": "color-mix(in srgb, var(--mantine-color-gray-light), transparent 80%)",
                        "marginBottom": "0.5rem",
                        "borderRadius": "0.25rem",
                    },
                    "content": {
                        "display": "flex",
                        "flexDirection": "column",
                        "gap": "0.375rem",
                        "padding": "0.125rem 0.5rem 0.5rem",
                    },
                },
                className=class_name,
            )
        elif self.render_type == "list":
            contents = dmc.Stack(
                [
                    self.list_item(
                        item=item,
                        aio_id=aio_id,
                        form_id=form_id,
                        field=field,
                        parent=parent,
                        index=i,
                        fields_repr=self.fields_repr,
                        sections=self.sections,
                        items_deletable=self.items_deletable,
                        read_only=self.read_only,
                    )
                    for i, _ in enumerate(value)
                ],
                id=self.ids.wrapper(aio_id, form_id, field, parent=parent),
                className=class_name,
            )
        elif self.render_type == "scalar":
            contents = html.Div(
                [
                    self.scalar_item(
                        item=item,
                        aio_id=aio_id,
                        form_id=form_id,
                        field=field,
                        parent=parent,
                        index=i,
                        fields_repr=self.fields_repr,
                        sections=self.sections,
                        items_deletable=self.items_deletable,
                        read_only=self.read_only,
                        input_kwargs=self.input_kwargs,
                    )
                    for i, _ in enumerate(value)
                ],
                id=self.ids.wrapper(aio_id, form_id, field, parent=parent),
                className=class_name,
                style={
                    "display": "grid",
                    "gridTemplateColumns": "repeat(auto-fit, minmax(min(100%, 280px), 1fr))",
                    "gap": "0.5rem",
                    "overflow": "hidden",
                    "alignItems": "top",
                },
            )
        elif self.render_type == "modal":
            contents = html.Div(
                [
                    self.modal_item(
                        item=item,
                        aio_id=aio_id,
                        form_id=form_id,
                        field=field,
                        parent=parent,
                        index=i,
                        value=val,
                        fields_repr=self.fields_repr,
                        sections=self.sections,
                        items_deletable=self.items_deletable,
                        read_only=self.read_only,
                    )
                    for i, val in enumerate(value)
                ],
                id=self.ids.wrapper(aio_id, form_id, field, parent=parent),
                style={
                    "display": "grid",
                    "gridTemplateColumns": "repeat(auto-fit, minmax(min(100%, 280px), 1fr))",
                    "gap": "0.5rem",
                    "overflow": "hidden",
                },
                className=class_name,
            )
        else:
            contents = self._contents_renderer(self.render_type)

        # Create a template item to be used clientside when adding new items
        template = self.render_type_item_mapper(self.render_type)(
            item=item.__class__.model_construct(),
            aio_id=aio_id,
            form_id=form_id,
            field=field,
            parent=parent,
            index="{{" + get_fullpath(parent, field).replace(":", "|") + "}}",
            value="-",
            opened=True,
            fields_repr=self.fields_repr,
            sections=self.sections,
            items_deletable=self.items_deletable,
            read_only=self.read_only,
            input_kwargs=self.input_kwargs,
        )
        title = self.get_title(field_info, field_name=field)
        description = self.get_description(field_info)

        return dmc.Stack(
            bool(title)
            * [
                dmc.Stack(
                    bool(title)
                    * [
                        dmc.Text(
                            [title]
                            + [
                                html.Span(
                                    " *", style={"color": "var(--input-asterisk-color, var(--mantine-color-error))"}
                                ),
                            ]
                            * self.is_required(field_info),
                            size="sm",
                            mt=3,
                            fw=500,
                            lh=1.55,
                        )
                    ]
                    + (bool(title) and bool(description)) * [dmc.Text(description, size="xs", c="dimmed", lh=1.2)],
                    gap=0,
                )
            ]
            + [
                contents,
                dcc.Store(
                    data=to_json_plotly(template), id=self.ids.template_store(aio_id, form_id, field, parent=parent)
                ),
            ]
            + self.items_creatable
            * [
                html.Div(
                    [
                        dmc.Button(
                            "Add",
                            leftSection=DashIconify(icon="carbon:add", height=16),
                            size="compact-md",
                            id=self.ids.add(aio_id, form_id, field, parent=parent),
                        ),
                    ],
                ),
            ],
            style={"gridColumn": "span var(--col-4-4)"},
            gap="0.5rem",
            mt="sm",
        )

    clientside_callback(
        ClientsideFunction(namespace="pydf", function_name="addToList"),
        Output(ids.wrapper(MATCH, MATCH, MATCH, MATCH), "children", allow_duplicate=True),
        Input(ids.add(MATCH, MATCH, MATCH, MATCH), "n_clicks"),
        State(ids.wrapper(MATCH, MATCH, MATCH, MATCH), "children"),
        State(ids.template_store(MATCH, MATCH, MATCH, MATCH), "data"),
        prevent_initial_call=True,
    )

    clientside_callback(
        ClientsideFunction(namespace="pydf", function_name="deleteFromList"),
        Output(ids.wrapper(MATCH, MATCH, MATCH, MATCH), "children", allow_duplicate=True),
        Input(ids.delete(MATCH, MATCH, MATCH, MATCH, ALL), "n_clicks"),
        State(ids.wrapper(MATCH, MATCH, MATCH, MATCH), "children"),
        prevent_initial_call=True,
    )

    # Open a model list modal when editing an item
    clientside_callback(
        ClientsideFunction(namespace="pydf", function_name="syncTrue"),
        Output(ids.modal(MATCH, MATCH, MATCH, MATCH, MATCH), "opened", allow_duplicate=True),
        Input(ids.edit(MATCH, MATCH, MATCH, MATCH, MATCH), "n_clicks"),
        prevent_initial_call=True,
    )

    # Close a model list modal when saving an item
    clientside_callback(
        ClientsideFunction(namespace="pydf", function_name="syncFalse"),
        Output(ids.modal(MATCH, MATCH, MATCH, MATCH, MATCH), "opened", allow_duplicate=True),
        Input(ids.modal_save(MATCH, MATCH, MATCH, MATCH, MATCH), "n_clicks"),
        prevent_initial_call=True,
    )

    # Update the modal title and list item to match the name field of the item (if it exists)
    clientside_callback(
        ClientsideFunction(namespace="pydf", function_name="updateModalTitle"),
        Output(ids.modal(MATCH, MATCH, "", MATCH, MATCH), "title"),
        Input(common_ids.value_field(MATCH, MATCH, "name", MATCH, MATCH), "value"),
        State(ids.modal(MATCH, MATCH, "", MATCH, MATCH), "id"),
    )

    # Update the accordion title to match the name field of the item (if it exists)
    clientside_callback(
        ClientsideFunction(namespace="pydf", function_name="updateAccordionTitle"),
        Output(ids.accordion_parent_text(MATCH, MATCH, "", MATCH, MATCH), "children"),
        Input(common_ids.value_field(MATCH, MATCH, "name", MATCH, MATCH), "value"),
    )



import json
from datetime import date, datetime
from enum import Enum
from typing import Literal

import dash_mantine_components as dmc
from dash import MATCH, Dash, Input, Output, State, _dash_renderer, callback, clientside_callback
from dash_iconify import DashIconify
from pydantic import BaseModel, Field, ValidationError

from dash_pydantic_form import FormSection, ModelForm, Sections, fields, get_model_cls, ids
from dash_pydantic_form.utils import SEP

_dash_renderer._set_react_version("18.2.0")
app = Dash(
    __name__,
    external_stylesheets=[
        "https://unpkg.com/@mantine/dates@7/styles.css",
        "https://unpkg.com/@mantine/code-highlight@7/styles.css",
        "https://unpkg.com/@mantine/charts@7/styles.css",
        "https://unpkg.com/@mantine/carousel@7/styles.css",
        "https://unpkg.com/@mantine/notifications@7/styles.css",
        "https://unpkg.com/@mantine/nprogress@7/styles.css",
    ],
)

server = app.server


class Species(Enum):
    """Species enum."""

    CAT = "cat"
    DOG = "dog"


class Pet(BaseModel):
    """Pet model."""

    name: str = Field(title="Name", description="Name of the pet")
    species: Species = Field(title="Species", description="Species of the pet")
    dob: date | None = Field(title="Date of birth", description="Date of birth of the pet", default=None)
    alive: bool = Field(title="Alive", description="Is the pet alive", default=True)

    def __str__(self) -> str:
        """String representation."""
        return str(self.name)


class Desk(BaseModel):
    """Desk model."""

    height: int = Field(title="Height", ge=0, repr_kwargs={"suffix": " cm"})
    material: str = Field(title="Material")


class WorkStation(BaseModel):
    """Work station model."""

    has_desk: bool = Field(title="Has desk")
    has_monitor: bool = Field(title="Has monitor")
    desk: Desk | None = Field(
        title="Desk",
        default=None,
        json_schema_extra={"repr_kwargs": {"visible": ("has_desk", "==", True)}},
    )


class HomeOffice(BaseModel):
    """Home office model."""

    type: Literal["home_office"] = Field(
        json_schema_extra={
            "repr_type": "RadioItems",
            "repr_kwargs": {"options_labels": {"home_office": "Home", "work_office": "Work"}},
        },
    )
    has_workstation: bool = Field(title="Has workstation", description="Does the employee have a suitable workstation")
    workstation: WorkStation | None = Field(
        title="Workstation",
        default=None,
        json_schema_extra={"repr_kwargs": {"visible": ("has_workstation", "==", True)}},
    )


class WorkOffice(BaseModel):
    """Work office model."""

    type: Literal["work_office"] = Field(
        json_schema_extra={
            "repr_type": "RadioItems",
            "repr_kwargs": {"options_labels": {"home_office": "Home", "work_office": "Work"}},
        },
    )
    commute_time: int = Field(title="Commute time", ge=0, repr_kwargs={"suffix": " min"})


class Metadata(BaseModel):
    """Metadata model."""

    languages: list[Literal["fr", "en", "sp", "cn"]] = Field(
        title="Languages spoken",
        default_factory=list,
        json_schema_extra={
            "repr_type": "Checklist",
            "repr_kwargs": {"options_labels": {"fr": "French", "en": "English", "sp": "Spanish", "cn": "Chinese"}},
        },
    )
    siblings: int | None = Field(title="Siblings", default=None, ge=0)
    location: HomeOffice | WorkOffice | None = Field(title="Work location", default=None, discriminator="type")


class Employee(BaseModel):
    """Employee model."""

    name: str = Field(title="Name", description="Name of the employee", min_length=2)
    age: int = Field(title="Age", description="Age of the employee, starting from their birth", ge=18)
    mini_bio: str | None = Field(
        title="Mini bio",
        description="Short bio of the employee",
        default=None,
        json_schema_extra={"repr_type": "Markdown"},
    )
    joined: datetime = Field(title="Joined", description="Date when the employee joined the company")
    office: Literal["au", "fr", "uk"] = Field(
        title="Office",
        description="Office of the employee",
        json_schema_extra={
            "repr_type": "RadioItems",
            "repr_kwargs": {"options_labels": {"au": "Australia", "fr": "France", "uk": "United Kingdom"}},
        },
    )
    metadata: Metadata | None = Field(title="Employee metadata", default=None)
    pets: list[Pet] = Field(title="Pets", description="Employee pets", default_factory=list)
    jobs: list[str] = Field(
        title="Past jobs", description="List of previous jobs the employee has held", default_factory=list
    )


bob = Employee(
    name="Bob",
    age=30,
    joined="2020-01-01",
    mini_bio="### Birth\nSomething something\n\n### Education\nCollege",
    office="au",
    metadata={
        "languages": ["fr", "en"],
        "siblings": 2,
        "location": {"type": "home_office", "has_workstation": True},
    },
    pets=[{"name": "Rex", "species": "cat"}],
    jobs=["Engineer", "Lawyer"],
)


AIO_ID = "home"
FORM_ID = "Bob"

app.layout = dmc.MantineProvider(
    id="mantine-provider",
    defaultColorScheme="auto",
    children=dmc.AppShell(
        [
            dmc.AppShellMain(
                dmc.Container(
                    [
                        dmc.Group(
                            [
                                dmc.Title("Dash Pydantic form", order=3),
                                dmc.Switch(
                                    offLabel=DashIconify(icon="radix-icons:moon", height=18),
                                    onLabel=DashIconify(icon="radix-icons:sun", height=18),
                                    size="md",
                                    color="yellow",
                                    persistence=True,
                                    checked=True,
                                    id="scheme-switch",
                                ),
                            ],
                            align="center",
                            justify="space-between",
                            mb="1rem",
                        ),
                        ModelForm(
                            Employee,
                            # bob,
                            AIO_ID,
                            FORM_ID,
                            # read_only=True,
                            # submit_on_enter=True,
                            # debounce_inputs=200,
                            fields_repr={
                                "name": {"placeholder": "Enter your name"},
                                "metadata": {
                                    "render_type": "accordion",
                                    "visible": ("_root_:office", "==", "au"),
                                },
                                "pets": fields.Table(
                                    fields_repr={
                                        "species": {"options_labels": {"dog": "Dog", "cat": "Cat"}},
                                    },
                                    table_height=200,
                                ),
                                "jobs": {"placeholder": "A job name"},
                            },
                            sections=Sections(
                                sections=[
                                    FormSection(name="General", fields=["name", "age", "mini_bio"], default_open=True),
                                    FormSection(
                                        name="HR",
                                        fields=["office", "joined", "location", "metadata"],
                                        default_open=True,
                                    ),
                                    FormSection(name="Other", fields=["pets", "jobs"], default_open=True),
                                ],
                                render="accordion",
                            ),
                        ),
                    ],
                    pt="1rem",
                )
            ),
            dmc.AppShellAside(
                dmc.ScrollArea(
                    dmc.Text(
                        id=ids.form_dependent_id("output", AIO_ID, FORM_ID),
                        style={"whiteSpace": "pre-wrap"},
                        p="1rem 0.5rem",
                    ),
                ),
            ),
        ],
        aside={"width": 350},
    ),
)


@callback(
    Output(ids.form_dependent_id("output", MATCH, MATCH), "children"),
    Output(ModelForm.ids.errors(MATCH, MATCH), "data"),
    Input(ModelForm.ids.main(MATCH, MATCH), "data"),
    State(ModelForm.ids.model_store(MATCH, MATCH), "data"),
    prevent_initial_call=True,
)
def display(form_data, model_name):
    """Display form data."""
    children = dmc.Stack(
        [
            dmc.Text("Form data", mb="-0.5rem", fw=600),
            dmc.Code(
                json.dumps(form_data, indent=2),
            ),
        ]
    )
    errors = None
    try:
        model_cls = get_model_cls(model_name)
        item = model_cls.model_validate(form_data)
        children.children[1].children = item.model_dump_json(indent=2)
    except ValidationError as e:
        children.children.extend(
            [
                dmc.Text("Validation errors", mb="-0.5rem", fw=500, c="red"),
                dmc.List(
                    [
                        dmc.ListItem(
                            [SEP.join([str(x) for x in error["loc"]]), f" : {error['msg']}, got {error['input']}"],
                        )
                        for error in e.errors()
                    ],
                    size="sm",
                    c="red",
                ),
            ]
        )
        errors = None
        errors = {SEP.join([str(x) for x in error["loc"]]): error["msg"] for error in e.errors()}

    return children, errors


clientside_callback(
    """(isLightMode) => isLightMode ? 'light' : 'dark'""",
    Output("mantine-provider", "forceColorScheme"),
    Input("scheme-switch", "checked"),
    prevent_initial_callback=True,
)


if __name__ == "__main__":
    app.run_server(debug=True)
